/** @file mqtt.c
 *  @brief Core handling of MQTT commands that may arrive from cloud.
 *
 */

#include "agg_appsupp.h"
#include "agg_iotsupp.h"

static void SendConfirmationCallback(IOTHUB_CLIENT_CONFIRMATION_RESULT result, void *userContextCallback);
static IOTHUBMESSAGE_DISPOSITION_RESULT ReceiveMessageCallback(IOTHUB_MESSAGE_HANDLE message, void *userContextCallback);

static void mqtt_process_vm_size(void *arg);

static const char* TAG = "MQTT";

static volatile bool dns_resolved;
static char propText[ MQTT_JSON_MAX_MSG_BYTE ];
char* reportedProperties;

static SemaphoreHandle_t MQTT_RESP;
struct timeval last_mqtt;   // Timestamp of last mqtt message sent to azure
struct timeval last_serial; // Timestamp of last message printed to serial
static int consecutive_send_failures;  // Keep track of the number of consecutive failed attempts to send mqtt message
int vmid;    // ID of a voice memo
int vmsize;  // Size of a voice memo

char *stun_ip;

//struct sockaddr_in servaddr;

IOTHUB_DEVICE_CLIENT_LL_HANDLE iotHubClientHandle;

static cloud_message_resp_t message_resp[] = {
  {.key = "VMSIZE",  .response_function = &mqtt_process_vm_size},
};


esp_err_t read_nvs_uint8(char *key, uint8_t *val)
{
    nvs_handle handle;
    esp_err_t err;
    err = nvs_open("storage", NVS_READONLY, &handle);
    if (err == ESP_OK)
    {
        err = nvs_get_u8(handle, key, val);
    }
    nvs_close(handle);
    return err;
}

esp_err_t write_nvs_uint8(char *key, uint8_t val)
{
    nvs_handle handle;
    esp_err_t err;
    err = nvs_open("storage", NVS_READWRITE, &handle);
    if (err == ESP_OK)
    {
        err = nvs_set_u8(handle, key, val);
        if (err == ESP_OK)
        {
            err = nvs_commit(handle);
        }
    }
    nvs_close(handle);
    return err;
}

void dns_found_cb(const char *name, const ip_addr_t *ipaddr, void *callback_arg)
{
    char *ip_str = inet_ntoa(*ipaddr);
    stun_ip = calloc(strlen(ip_str),sizeof(char));
    strcpy(stun_ip,ip_str);
    dns_resolved = true;
}

static void mqtt_process_vm_size(void *arg){
    ESP_LOGI(TAG,"mqtt_process_vm_size");

    vmsize = strtol((char *) arg,NULL,10);
}

void mqtt_init()
{  
    int receiveContext = 0;

    // Multiple instances of mqtt_message_response() task
    // might exist at the same time.  This mutex serializes
    // their use of the response functions.
    MQTT_RESP = xSemaphoreCreateMutex();

    // Azure port call to establish SNTP and
    // get first time sync.  May block for
    // for several secs.
 //   IoTHub_Init();

    uint8_t provisioned = false;
 //   esp_err_t err = read_nvs_u8( CA_NVS_DEV__U8_PROVIS, &provisioned );
    esp_err_t err = read_nvs_uint8("provisioned",&provisioned);

    printf("AAA AAA provis nvs = %d\n", provisioned );
 //   provisioned = true;  // AbCdEf

    if (err)
    {
        ESP_LOGE(TAG,"Reading provision flag from NVS failed");
    }
    else if (provisioned != false)
    {
        iotHubClientHandle = conn_dev_client();
    }

    if (iotHubClientHandle == NULL)
    {
        iotHubClientHandle = prov_dev_client(); //always provision device through certificates

        if (iotHubClientHandle == NULL)
        {
            fatal_error(TAG, "prov_dev_client failed after NVS string was not found in flash!");
        }
        provisioned = true;

     //   err = write_nvs_u8(CA_NVS_DEV__U8_PROVIS, provisioned);
        err = write_nvs_uint8("provisioned", provisioned);

        if (err)
        {
            ESP_LOGE(TAG, "Writing provision flag to NVS failed");
        }
    }

    srand((unsigned int)time(NULL));

    printf("BBB BBB returned from conn or prov dev client\n");

#if CONCISE_SERIAL == 0
    bool traceOn = true;

    IoTHubDeviceClient_LL_SetOption(iotHubClientHandle, "logtrace", &traceOn);
#endif

    IOTHUB_CLIENT_RESULT result;
  //  init_device_twin();
    if ((IoTHubDeviceClient_LL_SetMessageCallback(iotHubClientHandle, ReceiveMessageCallback, &receiveContext) != IOTHUB_CLIENT_OK))
    {
#if 0
        SetLedPattern( ALL_OFF );
        vTaskDelay(200);
        SetLedPattern( ONE_RED );
        vTaskDelay(200);
        SetLedPattern( ALL_OFF );
#endif        
    }
 //   if ((IoTHubDeviceClient_LL_SendReportedState(iotHubClientHandle, (const unsigned char *)reportedProperties, strlen(reportedProperties), reportedStateCallback, NULL) != IOTHUB_CLIENT_OK))
    {
#if 0
        SetLedPattern( ALL_OFF );
        vTaskDelay(200);
        SetLedPattern( ONE_RED );
        vTaskDelay(200);
        SetLedPattern( ALL_OFF );
#endif        
    }
 //   if ((IoTHubDeviceClient_LL_SetDeviceMethodCallback(iotHubClientHandle, deviceMethodCallback, NULL) != IOTHUB_CLIENT_OK))
    {
#if 0
        SetLedPattern( ALL_OFF );
        vTaskDelay(200);
        SetLedPattern( ONE_RED );
        vTaskDelay(200);
        SetLedPattern( ALL_OFF );
#endif    
    }
//    if ((IoTHubDeviceClient_LL_SetDeviceTwinCallback(iotHubClientHandle, deviceTwinCallback, &device_config) != IOTHUB_CLIENT_OK))
    {
#if 0
        SetLedPattern( ALL_OFF );
        vTaskDelay(200);
        SetLedPattern( ONE_RED );
        vTaskDelay(200);
        SetLedPattern( ALL_OFF );
#endif    
    }
    //IoTHubClient_LL_SetConnectionStatusCallback(iotHubClientHandle, ClientConnectionCallback, NULL);
}

void mqtt_deinit()
{
    IoTHubDeviceClient_LL_Destroy(iotHubClientHandle);
    platform_deinit();
}

void mqtt_message_response(void *properties)
{
    printf("RECV MSG RESPonse !!!\n");  // AbCdEf

    xSemaphoreTake(MQTT_RESP, 200/*portMAX_DELAY*/);

    MAP_HANDLE mapProperties = *(MAP_HANDLE *)properties;

    const char *const *keys;
    const char *const *values;
    size_t propertyCount = 0;
    if (mapProperties != NULL)
    {
        if (Map_GetInternals(mapProperties, &keys, &values, &propertyCount) == MAP_OK)
        {
            if (propertyCount > 0)
            {
                for (int i = 0; i < sizeof(message_resp) / sizeof(cloud_message_resp_t); i++)
                {
                    for (size_t index = 0; index < propertyCount; index++)
                    {
                        if (strcmp(keys[index], message_resp[i].key) == 0)
                        {
                            ESP_LOGI(TAG, "Key: %s Value: %s\r\n", keys[index], values[index]);
                            message_resp[i].response_function(values[index]);
                            break;
                        }
                    }
                }
            }
        }
    }
    Map_Destroy(mapProperties);
    xSemaphoreGive(MQTT_RESP);
    vTaskDelete(NULL);
}

static IOTHUBMESSAGE_DISPOSITION_RESULT ReceiveMessageCallback(IOTHUB_MESSAGE_HANDLE message, void *userContextCallback)
{
    printf("RECV MSG CALL BACK !!!\n");  // AbCdEf

    // Retrieve properties from the message
    MAP_HANDLE *mapProperties = malloc(sizeof(MAP_HANDLE));
    *mapProperties = Map_Clone(IoTHubMessage_Properties(message));
    if (mapProperties != NULL)
    {
        xTaskCreate(mqtt_message_response,"mqtt response",4096,mapProperties,11,NULL);
    }

    return IOTHUBMESSAGE_ACCEPTED;
}

static void SendConfirmationCallback(IOTHUB_CLIENT_CONFIRMATION_RESULT result, void *userContextCallback)
{
    // Possible IOTHUB_CLIENT_CONFIRMATION_RESULT result values: 
    //  IOTHUB_CLIENT_CONFIRMATION_OK,                   
    //  IOTHUB_CLIENT_CONFIRMATION_BECAUSE_DESTROY,      
    //  IOTHUB_CLIENT_CONFIRMATION_MESSAGE_TIMEOUT,      
    //  IOTHUB_CLIENT_CONFIRMATION_ERROR
  
    // Check result
    if (result != IOTHUB_CLIENT_CONFIRMATION_OK) {
        ESP_LOGE(TAG,"SendConfirmationCallback Failed With: 0x%x",(int)result);
     //   blank_leds();
     //   one_red();
        vTaskDelay(100);
     //   blank_leds();
        vTaskDelay(100);
        consecutive_send_failures+=1;
    }
    else
    {
        consecutive_send_failures = 0; //send was succesful, can reset counter
    }

    if(consecutive_send_failures > MQTT_MAX_FAIL)
    {
        fatal_error(TAG,"NUMBER OF CONSECUTIVE FAILED MQTT MESSAGE ATTEMPTS EXCEEDED");
    }
    

    EVENT_INSTANCE *eventInstance = (EVENT_INSTANCE *)userContextCallback;
    IoTHubMessage_Destroy(eventInstance->messageHandle);
}

void mqtt_do_work(void)
{
    // Only try if handle is initialized
    if ( iotHubClientHandle )
    {
        IoTHubDeviceClient_LL_DoWork(iotHubClientHandle);
    }
}
