/* WiFi station Example

   This example code is in the Public Domain (or CC0 licensed, at your option.)

   Unless required by applicable law or agreed to in writing, this
   software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
   CONDITIONS OF ANY KIND, either express or implied.
*/

#include "agg_appsupp.h"
#include "agg_iotsupp.h"

// Select legacy event loop scheme or newer scheme.
// Newer scheme will not work with v3.3 IDF.
#define USE_LEGACY_EVT_HDLR
#include "esp_event_loop.h"  // Needed to build with v3.3 IDF, although deprecated for v4.x

/* The examples use WiFi configuration that you can set via project configuration menu

   If you'd rather not, just change the below entries to strings with
   the config you want - ie #define EXAMPLE_WIFI_SSID "mywifissid"
*/
#define EXAMPLE_ESP_WIFI_SSID      CONFIG_ESP_WIFI_SSID
#define EXAMPLE_ESP_WIFI_PASS      CONFIG_ESP_WIFI_PASSWORD
#define EXAMPLE_ESP_MAXIMUM_RETRY  CONFIG_ESP_MAXIMUM_RETRY

#define WIFI_TIMEOUT_MSEC 10000    

/* FreeRTOS event group to signal when we are connected*/
static EventGroupHandle_t s_wifi_event_group;

static craftalert_conf_t g_device_config;  // Storage to hold device twin backed config
craftalert_conf_t *device_config = &g_device_config;   // device twin backed config
volatile state_t device_state;     // device state enum

uint8_t efuse_mac[BUFF_SZ_EFUSE_MAC];   // Persistent global to hold MAC raw bytes
char efuse_mac_str[BUFF_SZ_EFUSE_MAC_STR];

/* The event group allows multiple bits for each event, but we only care about one event 
 * - are we connected to the AP with an IP? */
const int WIFI_CONNECTED_BIT = BIT0;

static const char *TAG = "wifi station";

static int s_retry_num = 0;

// Use from other components
void fatal_error(const char * TAG, char * message) 
{
    ESP_LOGE(TAG,"%s\n",message);    
    vTaskDelay(6000);
    esp_restart();
}

#ifdef USE_LEGACY_EVT_HDLR
static esp_err_t event_handler(void *ctx, system_event_t *event)
{
    switch(event->event_id) {
    case SYSTEM_EVENT_STA_START:
        esp_wifi_connect();
        break;
    case SYSTEM_EVENT_STA_GOT_IP:
        ESP_LOGI(TAG, "got ip:%s",
                 ip4addr_ntoa(&event->event_info.got_ip.ip_info.ip));
        s_retry_num = 0;
        xEventGroupSetBits(s_wifi_event_group, WIFI_CONNECTED_BIT);
        break;
    case SYSTEM_EVENT_STA_DISCONNECTED:
        {
            if (s_retry_num < EXAMPLE_ESP_MAXIMUM_RETRY) {
                esp_wifi_connect();
                xEventGroupClearBits(s_wifi_event_group, WIFI_CONNECTED_BIT);
                s_retry_num++;
                ESP_LOGI(TAG,"retry to connect to the AP");
            }
            ESP_LOGI(TAG,"connect to the AP fail\n");
            break;
        }
    default:
        break;
    }
    return ESP_OK;
}
#else
// V4.x IDF and later only
static void event_handler(void* arg, esp_event_base_t event_base, 
                                int32_t event_id, void* event_data)
{
    if (event_base == WIFI_EVENT && event_id == WIFI_EVENT_STA_START) {
        esp_wifi_connect();
    } else if (event_base == WIFI_EVENT && event_id == WIFI_EVENT_STA_DISCONNECTED) {
        if (s_retry_num < EXAMPLE_ESP_MAXIMUM_RETRY) {
            esp_wifi_connect();
            xEventGroupClearBits(s_wifi_event_group, WIFI_CONNECTED_BIT);
            s_retry_num++;
            ESP_LOGI(TAG, "retry to connect to the AP");
        }
        ESP_LOGI(TAG,"connect to the AP fail");
    } else if (event_base == IP_EVENT && event_id == IP_EVENT_STA_GOT_IP) {
        ip_event_got_ip_t* event = (ip_event_got_ip_t*) event_data;
        ESP_LOGI(TAG, "got ip:%s",
                 ip4addr_ntoa(&event->ip_info.ip));
        s_retry_num = 0;
        xEventGroupSetBits(s_wifi_event_group, WIFI_CONNECTED_BIT);
    }
}
#endif

void wifi_init_sta()
{
    s_wifi_event_group = xEventGroupCreate();

    tcpip_adapter_init();

#ifdef USE_LEGACY_EVT_HDLR
    ESP_ERROR_CHECK(esp_event_loop_init(event_handler, NULL) );
#else
    // 4.0-beta2
    ESP_ERROR_CHECK(esp_event_loop_create_default());
#endif

    wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
    ESP_ERROR_CHECK(esp_wifi_init(&cfg));

#ifndef USE_LEGACY_EVT_HDLR
    // 4.0-beta2
    ESP_ERROR_CHECK(esp_event_handler_register(WIFI_EVENT, 
            ESP_EVENT_ANY_ID, &event_handler, NULL));
    ESP_ERROR_CHECK(esp_event_handler_register(IP_EVENT, 
            IP_EVENT_STA_GOT_IP, &event_handler, NULL));
#endif

    wifi_config_t wifi_config = {
        .sta = {
            .ssid = EXAMPLE_ESP_WIFI_SSID,
            .password = EXAMPLE_ESP_WIFI_PASS
        },
    };
    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_STA) );
    ESP_ERROR_CHECK(esp_wifi_set_config(ESP_IF_WIFI_STA, &wifi_config) );
    ESP_ERROR_CHECK(esp_wifi_start() );

    ESP_LOGI(TAG, "wifi_init_sta finished.");
    ESP_LOGI(TAG, "connect to ap SSID:%s password:%s",
             EXAMPLE_ESP_WIFI_SSID, EXAMPLE_ESP_WIFI_PASS);
}

void app_main()
{
    EventBits_t uxBits;

    //Initialize NVS
    esp_err_t ret = nvs_flash_init();
    if (ret == ESP_ERR_NVS_NO_FREE_PAGES || ret == ESP_ERR_NVS_NEW_VERSION_FOUND) {
      ESP_ERROR_CHECK(nvs_flash_erase());
      ret = nvs_flash_init();
    }
    ESP_ERROR_CHECK(ret);
    
    // Avoid spammy wifi logs
    esp_log_level_set("wpa", ESP_LOG_INFO);  
    esp_log_level_set("phy_init", ESP_LOG_INFO);  

    // Obtain raw MAC Address
    esp_efuse_mac_get_default(efuse_mac);

    // Format MAC so that it matches table format
    sprintf(efuse_mac_str, "%02x%02x%02x%02x%02x%02x", 
        efuse_mac[0], efuse_mac[1], efuse_mac[2], 
        efuse_mac[3], efuse_mac[4], efuse_mac[5]); 
 
    ESP_LOGI(TAG, "Efuse Mac %s", efuse_mac_str);

    // Init global device config struct.  Most accesses
    // during runtime use the pointer device_config.
    device_config->mac = efuse_mac_str; //uid_str;

    ESP_LOGI(TAG, "ESP_WIFI_MODE_STA");
    wifi_init_sta();

    // Now wait for completion notification before 
    // attempting SNTP, IoT provision.
    uxBits = xEventGroupWaitBits( s_wifi_event_group, WIFI_CONNECTED_BIT, 
   	      	pdTRUE/*clear on exit*/, pdFALSE/*wait on all*/, 
            pdMS_TO_TICKS( WIFI_TIMEOUT_MSEC ) );

	if (WIFI_CONNECTED_BIT & uxBits)
	{
		// Connection bit is set.
   	    ESP_LOGD(TAG, "STA Connect bit received.");

        IoTHub_Init();

   	    ESP_LOGW(TAG, "Attempt MQT INiT.");
        mqtt_init();
    }

    while (1)
    {
        vTaskDelay( 100 );
        printf("app_main wifi loop\n");
    }
}
